#include <windows.h>
#include <stdio.h>

struct DllClass;

int 
load_and_exec (const char *dllname, const char *procname, 
               const struct DllClass *dllclass) 
{
    HINSTANCE dllhandle;
    FARPROC proc;
    typedef int (*ifuncp) (const struct DllClass *dllclass);
    ifuncp dllfunc;
    int procfound = 1;
    int retcode;

    if (! (dllhandle = LoadLibrary(dllname))) {
	printf ("Error loading DLL `%s'\n", dllname);
	return 1;
    }

    if (! (proc = GetProcAddress (dllhandle, procname))) {
        procfound = 0;
    } 

    if (! procfound) {
	char* newname = (char *) malloc (strlen (procname) + 2);
        newname[0] = '_';
	strcpy (newname + 1, procname);
        if (! (proc = GetProcAddress (dllhandle, newname)))
          procfound = 0;
	else
          procfound = 1;
        free (newname);
    }

    if (! procfound) {
	printf ("Can't find proc `%s' (nor `_%s') in DLL `%s'\n", 
                procname, procname, dllname);
	return 1;
    }

    dllfunc = (ifuncp) proc;
    retcode = (*dllfunc) (dllclass);

    return retcode;
}

struct DllClass* 
new_dllclass (const char* dllname, const char *procname, int arg) {
    HINSTANCE dllhandle;
    FARPROC proc;
    typedef struct DllClass* (*ifuncp) (int);
    ifuncp dllfunc;
    int procfound = 1;
    struct DllClass* dllclass;

    if (! (dllhandle = LoadLibrary(dllname))) {
	printf ("Error loading DLL `%s'\n", dllname);
	return 0;
    }

    if (! (proc = GetProcAddress (dllhandle, procname))) {
        procfound = 0;
    } 

    if (! procfound) {
	char* newname = (char *) malloc (strlen (procname) + 2);
        newname[0] = '_';
	strcpy (newname + 1, procname);
        if (! (proc = GetProcAddress (dllhandle, newname)))
          procfound = 0;
	else
          procfound = 1;
        free (newname);
    }

    if (! procfound) {
	printf ("Can't find proc `%s' (nor `_%s') in DLL `%s'\n", 
                procname, procname, dllname);
	return 0;
    }

    dllfunc = (ifuncp) proc;
    dllclass = (*dllfunc) (arg);

    return dllclass;
}

void
delete_dllclass (const char* dllname, const char *procname, 
                 struct DllClass *dllclass) 
{
    HINSTANCE dllhandle;
    FARPROC proc;
    typedef void (*ifuncp) (struct DllClass *dllclass);
    ifuncp dllfunc;
    int procfound = 1;

    if (! (dllhandle = LoadLibrary(dllname))) {
	printf ("Error loading DLL `%s'\n", dllname);
	return;
    }

    if (! (proc = GetProcAddress (dllhandle, procname))) {
        procfound = 0;
    } 

    if (! procfound) {
	char* newname = (char *) malloc (strlen (procname) + 2);
        newname[0] = '_';
	strcpy (newname + 1, procname);
        if (! (proc = GetProcAddress (dllhandle, newname)))
          procfound = 0;
	else
          procfound = 1;
        free (newname);
    }

    if (! procfound) {
	printf ("Can't find proc `%s' (nor `_%s') in DLL `%s'\n", 
                procname, procname, dllname);
	return;
    }

    dllfunc = (ifuncp) proc;
    (*dllfunc) (dllclass);

    return;
}

int main (int argc, char *argv[]) {
    struct DllClass *dllclass;
    int retcode;

    dllclass = new_dllclass ("c++dll.dll", "new_dllclass", 5);
    retcode = load_and_exec ("c++dll.dll", "non_virtual_method", dllclass);
    retcode = load_and_exec ("c++dll.dll", "virtual_method", dllclass);
    delete_dllclass ("c++dll.dll", "delete_dllclass", dllclass);
    exit (0);
}

